"""Test of the jupyter plugin."""
from typing import Any
from typing import Dict
from typing import List

import pytest

from spectrafit.utilities.transformer import list2dict
from spectrafit.utilities.transformer import remove_none_type


@pytest.fixture
def reference_dict() -> Dict[str, Dict[str, Any]]:
    """Check reference dictionary.

    Returns:
        Dict[str, Dict[str, Any]]: Reference dictionary of two peaks.
    """
    return {
        "peaks": {
            "1": {
                "pseudovoigt": {
                    "amplitude": {"max": 2, "min": 0, "vary": True, "value": 1},
                    "center": {"max": 2, "min": -2, "vary": True, "value": 0},
                    "fwhmg": {"max": 0.1, "min": 0.02, "vary": True, "value": 0.01},
                    "fwhml": {"max": 0.1, "min": 0.01, "vary": True, "value": 0.01},
                }
            },
            "2": {
                "pseudovoigt": {
                    "amplitude": {"max": 2, "min": 0, "vary": True, "value": 1},
                    "center": {"max": 2, "min": -2, "vary": True, "value": 0},
                    "fwhmg": {"max": 0.1, "min": 0.02, "vary": True, "value": 0.01},
                    "fwhml": {"max": 0.1, "min": 0.01, "vary": True, "value": 0.01},
                }
            },
        },
    }


@pytest.fixture
def reference_list() -> List[Dict[str, Any]]:
    """Check reference list dictionary.

    Returns:
        Dict[str, Dict[str, Any]]: Reference list of dictionaries of two peaks.
    """
    return [
        {
            "pseudovoigt": {
                "amplitude": {"max": 2, "min": 0, "vary": True, "value": 1},
                "center": {"max": 2, "min": -2, "vary": True, "value": 0},
                "fwhmg": {"max": 0.1, "min": 0.02, "vary": True, "value": 0.01},
                "fwhml": {"max": 0.1, "min": 0.01, "vary": True, "value": 0.01},
            }
        },
        {
            "pseudovoigt": {
                "amplitude": {"max": 2, "min": 0, "vary": True, "value": 1},
                "center": {"max": 2, "min": -2, "vary": True, "value": 0},
                "fwhmg": {"max": 0.1, "min": 0.02, "vary": True, "value": 0.01},
                "fwhml": {"max": 0.1, "min": 0.01, "vary": True, "value": 0.01},
            }
        },
    ]


def test_converter(
    reference_list: List[Dict[str, Any]], reference_dict: Dict[str, Dict[str, Any]]
) -> None:
    """Test of the converter from list to dict.

    Args:
        reference_list (List[Dict[str, Any]]): Reference list of dictionaries of
             two peaks.
        reference_dict (Dict[str, Dict[str, Any]]): Reference dictionary of two peaks.
    """
    assert list2dict(reference_list) == reference_dict


def test_remove_none_type() -> None:
    """Test remove_none_type function."""
    d = {"a": None, "b": {"c": None, "d": {"e": None, "f": [1, None]}}}
    assert remove_none_type(d) == {"b": {"d": {"f": [1]}}}



"""Test of the Tools Model API."""

import pytest

from pydantic import ValidationError
from spectrafit.api.tools_model import AutopeakAPI


def test_raise_autopeak() -> None:
    """Test for raising exception of Autopeak Model."""
    with pytest.raises(ValidationError) as excinfo:
        AutopeakAPI(not_=2)  # type: ignore
    assert "not_" in str(excinfo.value)
    assert excinfo.type is ValidationError



"""Test of the notebook model."""

from _plotly_utils.colors.qualitative import Plotly as PlotlyColors
from spectrafit.api.notebook_model import ColorAPI
from spectrafit.api.notebook_model import FnameAPI
from spectrafit.api.notebook_model import FontAPI
from spectrafit.api.notebook_model import GridAPI
from spectrafit.api.notebook_model import LegendAPI
from spectrafit.api.notebook_model import MetricAPI
from spectrafit.api.notebook_model import PlotAPI
from spectrafit.api.notebook_model import ResidualAPI
from spectrafit.api.notebook_model import RunAPI
from spectrafit.api.notebook_model import XAxisAPI
from spectrafit.api.notebook_model import YAxisAPI


def test_xaxis_model() -> None:
    """Test the X-Axis model."""
    xaxis = XAxisAPI()
    assert xaxis.name == "Energy"
    assert xaxis.unit == "eV"


def test_yaxis_model() -> None:
    """Test the Y-Axis model."""
    yaxis = YAxisAPI()
    assert yaxis.name == "Intensity"
    assert yaxis.unit == "a.u."


def test_residual_model() -> None:
    """Test the Residual model."""
    residual = ResidualAPI()
    assert residual.name == "Residuals"
    assert residual.unit == "a.u."


def test_metric_model() -> None:
    """Test the Metric model."""
    metric = MetricAPI()
    assert metric.name_0 == "Metrics"
    assert metric.unit_0 == "a.u."
    assert metric.name_1 == "Metrics"
    assert metric.unit_1 == "a.u."


def test_run_model() -> None:
    """Test the Run model."""
    run = RunAPI()
    assert run.name == "Run"
    assert run.unit == "#"


def test_font_model() -> None:
    """Test the Font model."""
    font = FontAPI()
    assert font.family == "Open Sans, monospace"
    assert font.size == 12
    assert font.color == "black"


def test_legend_model() -> None:
    """Test the Legend model."""
    legend = LegendAPI()
    assert legend.orientation == "h"
    assert legend.x == 1
    assert legend.y == 1.02
    assert legend.xanchor == "right"
    assert legend.yanchor == "bottom"


def test_grid_model() -> None:
    """Test the Grid model."""
    grid = GridAPI()
    assert grid.show is True
    assert grid.ticks == "outside"
    assert grid.dash == "dot"


def test_fname_model() -> None:
    """Test the Fname model."""
    fname = FnameAPI(
        fname="test",
        suffix="png",
        prefix="test",
        folder="folder",
    )
    assert fname.fname == "test"
    assert fname.suffix == "png"
    assert fname.prefix == "test"
    assert fname.folder == "folder"


def test_plot_model() -> None:
    """Test the Plot model."""
    plot = PlotAPI(
        x="energy",
        y="intensity",
        title="test",
    )
    assert plot.x == "energy"
    assert plot.y == "intensity"
    assert plot.title == "test"
    assert isinstance(plot.xaxis_title, XAxisAPI)
    assert isinstance(plot.yaxis_title, YAxisAPI)


def test_color_model() -> None:
    """Test the Color model."""
    color = ColorAPI(paper="transparent")

    assert color.intensity == PlotlyColors[0]
    assert color.paper == "rgba(0,0,0,0)"



"""Transform the input data to a CSV file."""

import argparse
import re

from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pandas as pd

from spectrafit.api.file_model import DataFileAPI
from spectrafit.plugins.converter import Converter


def get_athena_column(fname: Path, comment: str = "#") -> Optional[List[str]]:
    """Get the header of the file.

    Args:
        fname (Path): The file name of the data file.
        comment (str, optional): The comment marker. Defaults to "#".

    Returns:
        Optional[List[str]]: The column names of the data file as a list.

    """
    with open(fname, encoding="utf-8") as f:
        data = f.read()
        lines = data.splitlines()
        return next(
            (
                lines[i - 1].split(comment)[-1].split()
                for i, line in enumerate(lines)
                if re.match(r"^\s*\d", line)
            ),
            None,
        )


athena_format = DataFileAPI(
    skiprows=None,
    skipfooter=0,
    delimiter=r"\s+",
    comment="#",
    names=get_athena_column,
    file_suffixes=[".nor"],
)


txt_format = DataFileAPI(
    skiprows=None,
    skipfooter=0,
    delimiter=r"\s+",
    header=0,
    file_suffixes=[".txt", ".dat", ".out"],
)


@dataclass(frozen=True)
class DataFormats:
    """Data formats."""

    ATHENA = athena_format
    TXT = txt_format


# Remove the private attributes
choices = {
    choice for choice in list(DataFormats.__dict__.keys()) if not choice.startswith("_")
}

choices_export = {"csv", "txt", "dat", "out"}


class DataConverter(Converter):
    """Convert the data files to a CSV file.

    !!! info "Supported file formats"

        Currently supported file formats:

        -[x] ATHENA
        -[x] TXT
        -[ ] more to come

        `DataConverter` class can be also used in the Jupyter notebook.
    """

    def get_args(self) -> Dict[str, Any]:
        """Get the arguments from the command line.

        Returns:
            Dict[str, Any]: Return the input file arguments as a dictionary without
                additional information beyond the command line arguments.
        """
        parser = argparse.ArgumentParser(
            description="Converter for 'SpectraFit' from data files to CSV files."
        )
        parser.add_argument(
            "infile",
            type=Path,
            help="Filename of the data file to convert.",
        )
        parser.add_argument(
            "-f",
            "--file-format",
            help="File format for the conversion.",
            type=str,
            choices=choices,
        )
        parser.add_argument(
            "-e",
            "--export-format",
            help="File format for the export.",
            type=str,
            default="csv",
            choices=choices_export,
        )
        return vars(parser.parse_args())

    @staticmethod
    def convert(infile: Path, file_format: str) -> pd.DataFrame:
        """Convert the input file to the target file format.

        Args:
            infile (Path): Input file as a path object.
            file_format (str): Target file format.

        Raises:
            ValueError: If the file format is not supported.

        Returns:
            pd.DataFrame: The converted data as a pandas DataFrame.
        """
        if file_format.upper() not in choices:
            raise ValueError(f"File format '{file_format}' is not supported.")

        if callable(DataFormats.__dict__[file_format].names):
            names = DataFormats.__dict__[file_format].names(infile)
        else:
            names = DataFormats.__dict__[file_format].names
        DataFormats.__dict__[file_format].names = names

        return pd.read_csv(
            infile, **DataFormats.__dict__[file_format].dict(exclude={"file_suffixes"})
        )

    def save(self, data: Any, fname: Path, export_format: str) -> None:
        """Save the converted data to a CSV file.

        Raises:
            ValueError: If the export format is not supported.

        Args:
            data (Any): The converted data, which is a pandas DataFrame.
            fname (Path): The file name of the data file.
            export_format (str): The file format of the exported file.
        """
        if export_format.lower() not in choices_export:
            raise ValueError(f"Export format '{export_format}' is not supported.")
        data.to_csv(fname.with_suffix(f".{export_format}"), index=False)

    def __call__(self) -> None:
        """Run the converter."""
        args = self.get_args()
        self.save(
            data=self.convert(
                args["infile"],
                args["file_format"],
            ),
            fname=args["infile"],
            export_format=args["export_format"],
        )


def command_line_runner() -> None:
    """Run the converter from the command line."""
    DataConverter()()



"""Reference model for the API of the Jupyter Notebook report."""
from typing import Any
from typing import Dict
from typing import Hashable
from typing import List
from typing import Union

from dtale import __version__ as dtale_version
from emcee import __version__ as emcee_version
from itables import __version__ as itables_version
from lmfit import __version__ as lmfit_version
from numdifftools import __version__ as numdifftools_version
from numpy import __version__ as numpy_version
from pandas import __version__ as pandas_version
from plotly import __version__ as plotly_version
from pydantic import BaseModel
from pydantic import Field
from pydantic import __version__ as pydantic_version
from scipy import __version__ as scipy_version
from sklearn import __version__ as sklearn_version
from spectrafit.api.cmd_model import DescriptionAPI
from spectrafit.api.tools_model import DataPreProcessingAPI
from statsmodels import __version__ as statsmodels_version


class CreditsAPI(BaseModel, allow_mutation=False):
    """Credits API model."""

    dtale: str = f"dtale v{dtale_version}"
    emcee: str = f"emcee v{emcee_version}"
    itables: str = f"itables v{itables_version}"
    lmfit: str = f"lmfit v{lmfit_version}"
    numdifftools: str = f"numdifftools v{numdifftools_version}"
    numpy: str = f"numpy v{numpy_version}"
    pandas: str = f"pandas v{pandas_version}"
    plotly: str = f"plotly v{plotly_version}"
    pydantic: str = f"pydantic v{pydantic_version}"
    scipy: str = f"scipy v{scipy_version}"
    sklearn: str = f"sklearn v{sklearn_version}"
    statsmodels: str = f"statsmodels v{statsmodels_version}"


class FitMethodAPI(BaseModel):
    """Fit method API model."""

    global_fitting: Union[bool, int] = Field(
        default=False,
        description="Fitting in the global fashion",
    )
    confidence_interval: Union[bool, Dict[str, Any]] = Field(
        ...,
        description="Settings for the confidence interval calculation",
    )
    configurations: Dict[str, Any] = Field(
        ..., description="Settings for the fitting configuration"
    )
    settings_solver_models: Dict[str, Any] = Field(
        ...,
        description="Settings for the solver models including minimizer and optimizer",
    )


class InputAPI(BaseModel):
    """Input API for the report endpoint."""

    description: DescriptionAPI = DescriptionAPI()
    credits: CreditsAPI = CreditsAPI()
    initial_model: List[Dict[str, Dict[str, Dict[str, Any]]]] = Field(
        ..., description="Initial model for the fit"
    )
    method: FitMethodAPI = Field(
        ..., description="Fitting method with optional including of confidence interval"
    )
    pre_processing: DataPreProcessingAPI = Field(..., description="Data pre-processing")


class SolverAPI(BaseModel):
    """Solver API for the report endpoint."""

    goodness_of_fit: Dict[str, float] = Field(..., description="Goodness of fit")
    regression_metrics: Dict[str, List[Any]] = Field(
        ..., description="Regression metrics"
    )
    descriptive_statistic: Dict[str, List[Any]] = Field(
        ..., description="Descriptive statistic"
    )
    linear_correlation: Dict[str, List[Any]] = Field(
        ..., description="Linear correlation"
    )
    component_correlation: Dict[str, Dict[str, Any]] = Field(
        default={},
        description="Linear correlation of each attribute of components. if possible",
    )
    confidence_interval: Dict[str, Any] = Field(
        default={}, description="Confidence interval, if possible"
    )
    covariance_matrix: Dict[str, Dict[str, Any]] = Field(
        default={}, description="Covariance matrix, if possible"
    )
    variables: Dict[str, Dict[str, Any]] = Field(
        ...,
        description="Variables with their initial, optimized and optional error values",
    )
    errorbars: Dict[str, Any] = Field(
        default={},
        description="Error bar comment if values reach initial value or boundary",
    )


class OutputAPI(BaseModel):
    """Output API for the report endpoint."""

    df_org: Dict[Hashable, Any] = Field(
        ...,
        description="DataFrame of the original data via 'records' orient",
    )
    df_fit: Dict[Hashable, Any] = Field(
        ...,
        description="DataFrame of the fitted data via 'records' orient",
    )
    df_pre: Dict[Hashable, Any] = Field(
        default={},
        description="DataFrame of the pre-processed data via 'records' orient",
    )

    class Config:
        """Config for the OutputAPI of arbitary types."""

        arbitrary_types_allowed = True


class ReportAPI(BaseModel):
    """Definition of the report model."""

    input: InputAPI = Field(
        ...,
        description="Input data for the report.",
    )
    solver: SolverAPI = Field(
        ...,
        description="Solver data for the report.",
    )
    output: OutputAPI = Field(
        ...,
        description="Output data for the report.",
    )



"""This module contains the RIXS visualizer class."""

import argparse
import json
import logging

from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import dash
import dash_bootstrap_components as dbc
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import tomli

from dash import dcc
from dash import html
from dash_bootstrap_templates import ThemeChangerAIO
from dash_bootstrap_templates import template_from_url
from jupyter_dash import JupyterDash
from numpy.typing import NDArray
from spectrafit.api.rixs_model import MainTitleAPI
from spectrafit.api.rixs_model import RIXSModelAPI
from spectrafit.api.rixs_model import SizeRatioAPI
from spectrafit.api.rixs_model import XAxisAPI
from spectrafit.api.rixs_model import YAxisAPI
from spectrafit.api.rixs_model import ZAxisAPI
from spectrafit.plugins.notebook import DataFramePlot


class RIXSFigure:
    """Class to create the RIXS figure.

    !!! info "About the RIXS figure"

        The RIXS figure is composed of three subplots:

        - RIXS -> 3D plot
        - XES -> 2D plot
        - XAS -> 2D plot

    """

    def __init__(
        self,
        incident_energy: NDArray[np.float64],
        emission_energy: NDArray[np.float64],
        rixs_map: NDArray[np.float64],
        size: SizeRatioAPI = SizeRatioAPI(
            size=(500, 500),
            ratio_rixs=(2, 2),
            ratio_xes=(3, 1),
            ratio_xas=(3, 1),
        ),
        x_axis: XAxisAPI = XAxisAPI(name="Incident Energy", unit="eV"),
        y_axis: YAxisAPI = YAxisAPI(name="Emission Energy", unit="eV"),
        z_axis: ZAxisAPI = ZAxisAPI(name="Intensity", unit="a.u."),
    ):
        """Initialize the RIXS figure.

        Args:
            incident_energy (NDArray[np.float64]): Incident energy.
            emission_energy (NDArray[np.float64]): Emission energy.
            rixs_map (NDArray[np.float64]): RIXS data as a 2D array.
            size (SizeRatioAPI, optional): Size of the figure.
                 Defaults to SizeRatioAPI(size=(500, 500), ratio_rixs=(2, 2),
                 ratio_xes=(3, 1), ratio_xas=(3, 1)).
            x_axis (XAxisAPI, optional): X-Axis of the figure.
                 Defaults to XAxisAPI(name="Incident Energy", unit="eV").
            y_axis (YAxisAPI, optional): Y-Axis of the figure.
                 Defaults to YAxisAPI(name="Emission Energy", unit="eV").
            z_axis (ZAxisAPI, optional): Z-Axis of the figure.
                 Defaults to ZAxisAPI(name="Intensity", unit="a.u.").
        """
        self.incident_energy = incident_energy
        self.emission_energy = emission_energy
        self.rixs_map = rixs_map

        self.x_axis = x_axis
        self.y_axis = y_axis
        self.z_axis = z_axis
        self.initialize_figure_size(size)

    def initialize_figure_size(self, size: SizeRatioAPI) -> None:
        """Initialize the size of the figure.

        Args:
            size (SizeRatioAPI): Size of the figure.
        """
        self.rixs_width = int(size.size[0] * size.ratio_rixs[0])
        self.rixs_height = int(size.size[1] * size.ratio_rixs[1])
        self.xas_width = int(size.size[0] * size.ratio_xas[0])
        self.xas_height = int(size.size[1] * size.ratio_xas[1])
        self.xes_width = int(size.size[0] * size.ratio_xes[0])
        self.xes_height = int(size.size[1] * size.ratio_xes[1])

    def create_rixs(
        self,
        colorscale: str = "Viridis",
        opacity: float = 0.9,
        template: Optional[str] = None,
    ) -> go.Figure:
        """Create the RIXS figure.

        Args:
            colorscale (str, optional): Color scale. Defaults to "Viridis".
            opacity (float, optional): Opacity of the surface. Defaults to 0.9.
            template (str, optional): Template of the figure. Defaults to None.

        Returns:
            go.Figure: RIXS figure.
        """
        fig = go.Figure(
            data=[
                go.Surface(
                    x=self.incident_energy,
                    y=self.emission_energy,
                    z=self.rixs_map,
                    colorscale=colorscale,
                    opacity=opacity,
                    contours_z=dict(
                        show=True,
                        usecolormap=True,
                        highlightcolor="limegreen",
                        project_z=True,
                    ),
                )
            ],
        )

        fig.update_layout(
            autosize=True,
            width=self.rixs_width,
            height=self.rixs_height,
            scene=dict(
                xaxis_title=DataFramePlot.title_text(
                    name=self.x_axis.name, unit=self.x_axis.unit
                ),
                yaxis_title=DataFramePlot.title_text(
                    name=self.y_axis.name, unit=self.y_axis.unit
                ),
                zaxis_title=DataFramePlot.title_text(
                    name=self.z_axis.name, unit=self.z_axis.unit
                ),
                aspectmode="cube",
            ),
            template=template,
        )
        fig.update_traces(
            contours_z=dict(
                show=True, usecolormap=True, highlightcolor="limegreen", project_z=True
            )
        )
        return fig

    def create_xes(
        self,
        x: NDArray[np.float64],
        y: NDArray[np.float64],
        template: Optional[str] = None,
    ) -> go.Figure:
        """Create the XES figure.

        Args:
            x (NDArray[np.float64]): X-axis of the figure.
            y (NDArray[np.float64]): Y-axis of the figure.
            template (str, optional): Template of the figure. Defaults to None.

        Returns:
            go.Figure: XES figure.
        """
        fig = px.line(x=x, y=y, template=template)
        fig.update_layout(
            autosize=True,
            width=self.xes_width,
            height=self.xes_height,
        )
        # Udate the xaxis title
        fig.update_xaxes(
            title_text=DataFramePlot.title_text(
                name=self.y_axis.name, unit=self.y_axis.unit
            )
        )
        # Update the yaxis title
        fig.update_yaxes(
            title_text=DataFramePlot.title_text(
                name=self.z_axis.name, unit=self.z_axis.unit
            )
        )
        return fig

    def create_xas(
        self,
        x: NDArray[np.float64],
        y: NDArray[np.float64],
        template: Optional[str] = None,
    ) -> go.Figure:
        """Create the XAS figure.

        Args:
            x (NDArray[np.float64]): X-axis of the figure.
            y (NDArray[np.float64]): Y-axis of the figure.
            template (str, optional): Template of the figure. Defaults to None.

        Returns:
            go.Figure: XAS figure.
        """
        fig = px.line(x=x, y=y, template=template)
        fig.update_layout(
            autosize=True,
            width=self.xas_width,
            height=self.xas_height,
        )
        fig.update_xaxes(
            title_text=DataFramePlot.title_text(
                name=self.x_axis.name, unit=self.x_axis.unit
            )
        )
        fig.update_yaxes(
            title_text=DataFramePlot.title_text(
                name=self.z_axis.name, unit=self.z_axis.unit
            )
        )
        return fig


class RIXSApp(RIXSFigure):  # pragma: no cover
    """Create the RIXS app.

    !!! info "About the RIXS app"

        The RIXS app is a web application that allows you to visualize the RIXS
        data. The app is based on the Dash framework. The app is composed of
        three figures: the RIXS figure, the XES figure and the XAS figure.

        The RIXS figure is a 3D surface plot. The XES figure is a line plot
        showing the XES spectrum. The XAS figure is a line plot showing the XAS
        spectrum.

        The RIXS figure is interactive. You can zoom in and out, rotate the
        figure, and change the color scale. The XES and XAS figures are not
        interactive.

    """

    def __init__(
        self,
        incident_energy: NDArray[np.float64],
        emission_energy: NDArray[np.float64],
        rixs_map: NDArray[np.float64],
        size: SizeRatioAPI = SizeRatioAPI(
            size=(500, 500),
            ratio_rixs=(2, 2),
            ratio_xas=(3, 1),
            ratio_xes=(3, 1),
        ),
        main_title: MainTitleAPI = MainTitleAPI(rixs="RIXS", xes="XES", xas="XAS"),
        fdir: Path = Path("./"),
        mode: str = "server",
        jupyter_dash: bool = False,
        port: int = 8050,
        debug: bool = False,
    ) -> None:
        """Create the RIXS app.

        Args:
            incident_energy (NDArray[np.float64]): Incident energy.
            emission_energy (NDArray[np.float64]): Emission energy.
            rixs_map (NDArray[np.float64]): RIXS data as a 2D array.
            size (SizeRatioAPI, optional): Size of the figures. Defaults to
                 SizeRatioAPI(size=(500, 500), ratio_rixs=(2, 2), ratio_xas=(3, 1),
                 ratio_xes=(3, 1)).
            main_title (MainTitleAPI, optional): Main title of the figures.
                 Defaults to MainTitleAPI(rixs="RIXS", xes="XES", xas="XAS").
            fdir (Path, optional): Directory to save the figures. Defaults to
                 Path("./").
            mode (str, optional): Mode of the app. Defaults to "server".
            port (int, optional): Port of the app. Defaults to 8050.
            jupyter_dash (bool, optional): Jupyter Dash mode. Defaults to False.
            debug (bool, optional): Debug mode. Defaults to False.

        """
        super().__init__(
            incident_energy=incident_energy,
            emission_energy=emission_energy,
            rixs_map=rixs_map,
            size=size,
        )
        self.fdir = fdir
        self.main_title = main_title
        self.mode = mode
        self.jupyter_dash = jupyter_dash
        self.port = port
        self.debug = debug
        if not self.debug:
            self.logging_flask()

    def logging_flask(self) -> None:
        """Set the logging level of the Flask server to ERROR."""
        log = logging.getLogger("werkzeug")
        log.setLevel(logging.ERROR)

    def colorscale(self) -> html.Div:
        """Create the color scale dropdown.

        Returns:
            html.Div: Color scale dropdown.
        """
        return html.Div(
            [
                dbc.Label("Color Scale"),
                dcc.Dropdown(
                    id="colorscale",
                    options=[
                        {"label": "Viridis", "value": "Viridis"},
                        {"label": "Plasma", "value": "Plasma"},
                        {"label": "Inferno", "value": "Inferno"},
                        {"label": "Magma", "value": "Magma"},
                        {"label": "Cividis", "value": "Cividis"},
                        {"label": "Greys", "value": "Greys"},
                        {"label": "Greens", "value": "Greens"},
                        {"label": "YlOrRd", "value": "YlOrRd"},
                        {"label": "Bluered", "value": "Bluered"},
                        {"label": "RdBu", "value": "RdBu"},
                        {"label": "Reds", "value": "Reds"},
                        {"label": "Blues", "value": "Blues"},
                        {"label": "Picnic", "value": "Picnic"},
                        {"label": "Rainbow", "value": "Rainbow"},
                        {"label": "Portland", "value": "Portland"},
                        {"label": "Jet", "value": "Jet"},
                        {"label": "Hot", "value": "Hot"},
                        {"label": "Blackbody", "value": "Blackbody"},
                        {"label": "Earth", "value": "Earth"},
                        {"label": "Electric", "value": "Electric"},
                        {"label": "Viridis", "value": "Viridis"},
                        {"label": "Cividis", "value": "Cividis"},
                    ],
                    value="Viridis",
                ),
            ],
            className="dbc",
        )

    def opacity(self) -> html.Div:
        """Create the opacity slider.

        Returns:
            html.Div: Opacity slider.
        """
        return html.Div(
            [
                dbc.Label("Opacity"),
                dcc.Slider(
                    id="opacity",
                    min=0,
                    max=1,
                    step=0.1,
                    value=1,
                    marks={i: str(i) for i in range(2)},
                ),
            ]
        )

    def header(self) -> dbc.Card:
        """Create the header.

        Returns:
            dbc.Card: Header as a bootstrap card.
        """
        return dbc.Card(
            dbc.CardBody(
                [
                    html.H4(
                        "RIXS Visualizer App",
                        className="bg-primary text-white p-2 mb-2 text-center",
                    )
                ]
            )
        )

    def pre_body(self) -> Tuple[html.Div, html.Div, html.Div]:
        """Create the body.

        Returns:
            Tuple[html.Div, html.Div, html.Div]: Body as a tuple of three plot parts.
        """
        rixs = html.Div(
            [
                dbc.Label(self.main_title.rixs),
                dcc.Graph(id="rixs-figure"),
            ]
        )
        xes = html.Div(
            [
                dbc.Label(self.main_title.xes),
                dcc.Graph(id="xes-figure"),
            ]
        )
        xas = html.Div(
            [
                dbc.Label(self.main_title.xas),
                dcc.Graph(id="xas-figure"),
            ]
        )
        return rixs, xes, xas

    def body(self) -> dbc.Card:
        """Create the body.

        Returns:
            dbc.Card: Body as a bootstrap card.
        """
        colorscale = self.colorscale()
        opacity = self.opacity()
        rixs, xes, xas = self.pre_body()

        return (
            dbc.Card(
                dbc.CardBody(
                    [
                        dbc.Row([ThemeChangerAIO(aio_id="theme")]),
                        dbc.Row(
                            [
                                dbc.Col(
                                    html.H1("RIXS Viewer", className="text-center")
                                ),
                            ],
                            justify="left",
                        ),
                        html.Br(),
                        dbc.Row(
                            [
                                dbc.Col(colorscale),
                                dbc.Col(opacity),
                            ],
                        ),
                        html.Br(),
                        dbc.Row(
                            [
                                dbc.Col(rixs),
                                dbc.Col([xes, xas]),
                            ],
                            justify="left",
                        ),
                        html.Br(),
                    ],
                ),
                class_name="mt-4",
            ),
        )[0]

    def footer(self) -> dbc.Card:
        """Create the footer.

        Returns:
            dbc.Card: Footer as a bootstrap card.
        """
        return (
            dbc.Card(
                dbc.CardBody(
                    [
                        dbc.Row(
                            [
                                dcc.Markdown(
                                    """
                    ### RIXS Viewer
                    This is a simple RIXS viewer. It is based on the
                    [Dash](https://dash.plotly.com/)
                    framework and uses the [Plotly](https://plotly.com/python/) library
                    for plotting. The code is available on
                    [GitHub](https://github.com/anselmoo/spectrafit).
                    """
                                ),
                            ],
                            justify="left",
                        )
                    ]
                ),
                class_name="mt-4",
            ),
        )[0]

    def app_run(self) -> None:
        """Run the app."""
        dbc_css = (
            "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.min.css"
        )
        external_stylesheets = [dbc.themes.COSMO, dbc_css]
        if self.jupyter_dash:
            app = JupyterDash(__name__, external_stylesheets=external_stylesheets)
        else:
            app = dash.Dash(
                __name__,
                external_stylesheets=external_stylesheets,
                meta_tags=[
                    {
                        "name": "viewport",
                        "content": "width=device-width, initial-scale=1",
                    }
                ],
            )
        app.layout = dbc.Container(
            [
                self.header(),
                self.body(),
                self.footer(),
            ],
            fluid=True,
        )

        @app.callback(
            [
                dash.dependencies.Output("xes-figure", "figure"),
                dash.dependencies.Output("xas-figure", "figure"),
                dash.dependencies.Output("rixs-figure", "figure"),
            ],
            [
                dash.dependencies.Input("rixs-figure", "hoverData"),
                dash.dependencies.Input("rixs-figure", "clickData"),
                dash.dependencies.Input("colorscale", "value"),
                dash.dependencies.Input("opacity", "value"),
                dash.dependencies.Input(ThemeChangerAIO.ids.radio("theme"), "value"),
            ],
        )
        def update_hover_data(
            hoverData: Dict[str, List[Dict[str, float]]],
            clickData: Dict[str, List[Dict[str, float]]],
            colorscale: str,
            opacity: float,
            theme: str,
        ) -> Tuple[go.Figure, go.Figure, go.Figure]:
            if hoverData is None:
                return (
                    self.create_xas(
                        x=self.incident_energy,
                        y=self.rixs_map[:, int(self.emission_energy.size / 2)],
                        template=template_from_url(theme),
                    ),
                    self.create_xes(
                        x=self.emission_energy,
                        y=self.rixs_map[int(self.incident_energy.size / 2), :],
                        template=template_from_url(theme),
                    ),
                    self.create_rixs(
                        colorscale=colorscale,
                        opacity=opacity,
                        template=template_from_url(theme),
                    ),
                )
            x = hoverData["points"][0]["x"]
            y = hoverData["points"][0]["y"]
            xes_fig = self.create_xas(
                x=self.incident_energy,
                y=self.rixs_map[:, int(x)],
                template=template_from_url(theme),
            )
            xas_fig = self.create_xes(
                x=self.emission_energy,
                y=self.rixs_map[int(y), :],
                template=template_from_url(theme),
            )
            rixs_fig = self.create_rixs(
                colorscale=colorscale,
                opacity=opacity,
                template=template_from_url(theme),
            )
            if clickData is None:
                return xes_fig, xas_fig, rixs_fig
            cx = clickData["points"][0]["x"]
            cy = clickData["points"][0]["y"]
            pd.DataFrame(
                {"energy": self.emission_energy, "intensity": self.rixs_map[int(cy), :]}
            ).to_csv(
                self.fdir / f"xes_cut_{np.round(cx, 8)}.txt",
                index=False,
            )
            pd.DataFrame(
                {"energy": self.incident_energy, "intensity": self.rixs_map[:, int(cx)]}
            ).to_csv(
                self.fdir / f"xas_cut_{np.round(cy, 8)}.txt",
                index=False,
            )
            return xes_fig, xas_fig, rixs_fig

        if self.jupyter_dash:
            app.run_server(mode=self.mode, debug=self.debug, port=self.port)
        else:
            app.run_server(debug=self.debug, port=self.port)


class RIXSVisualizer:
    """RIXS Visualizer. This class is used to visualize RIXS data."""

    def get_args(self) -> Dict[str, Any]:
        """Get the arguments from the command line.

        Returns:
            Dict[str, Any]: Return the input file arguments as a dictionary without
                additional information beyond the command line arguments.
        """
        parser = argparse.ArgumentParser(
            description="`RIXS-Visualizer` is a simple RIXS plane viewer, which "
            "allows to visualize RIXS data in a 2D plane."
        )
        parser.add_argument(
            "infile",
            type=Path,
            help="The input file. This can be a json, toml, npy, or npz file.",
        )
        return vars(parser.parse_args())

    @staticmethod
    def load_data(infile: Path) -> RIXSModelAPI:
        """Load the data from the input file.

        Args:
            infile (Path): The input file path. This can be a json, toml, npy, or npz
                file.

        Raises:
            ValueError: If the file type is not supported.

        Returns:
            RIXSModelAPI: The data as a pydantic model object with the following
                attributes: incident_energy, emission_energy, and rixs_map. The
                incident_energy and emission_energy are 1D arrays, and the rixs_map is
                a 2D array.
        """
        if infile.suffix == ".npy":
            data = np.load(infile, allow_pickle=True).item()
        elif infile.suffix == ".npz":
            data = np.load(infile, allow_pickle=True)
        elif infile.suffix == ".json":
            with open(infile, encoding="utf-8") as f:
                data = json.load(f)
        elif infile.suffix in {".toml", ".lock"}:
            with open(infile, "rb") as f:
                data = tomli.load(f)
        else:
            raise ValueError(f"File type {infile.suffix} is not supported.")
        return RIXSModelAPI(
            incident_energy=np.array(data["incident_energy"]),
            emission_energy=np.array(data["emission_energy"]),
            rixs_map=np.array(data["rixs_map"]),
        )

    def __call__(self) -> None:  # pragma: no cover
        """Run the RIXS Visualizer."""
        app = RIXSApp(**self.load_data(self.get_args()["infile"]).dict())
        app.app_run()


def command_line_runner() -> None:
    """Run the RIXS Visualizer from the command line."""
    RIXSVisualizer()()



"""Test the model model."""

import pytest

from spectrafit.api.models_model import DistributionModelAPI


@pytest.mark.parametrize(
    "distribution_model",
    [
        "gaussian",
        "lorentzian",
        "voigt",
        "pseudovoigt",
        "exponential",
        "power",
        "linear",
        "constant",
        "erf",
        "heaviside",
        "atan",
        "log",
        "cgaussian",
        "clorentzian",
        "cvoigt",
    ],
)
def test_distribution_model(distribution_model: str) -> None:
    """Test the distribution model."""
    assert distribution_model in list(
        DistributionModelAPI.schema()["properties"].keys()
    )
